
-- http://www.modbus.org/docs/Modbus_Messaging_Implementation_Guide_V1_0b.pdf
-- http://www.modbus.org/docs/Modbus_Application_Protocol_V1_1b.pdf


--require "socket"
require "struct"
require "tools"

-- some settings
CONNECT_TIMEOUT = 15
RECONNECT_INTERVAL = 5
REQUEST_TIMEOUT = 30


-- error codes
ERROR_TIMEOUT    = 901
ERROR_DATA       = 902
ERROR_CONNECTION = 903


-----------------------------------------------------------------------------
-- ModBus addresses range
class "AddrRange"

-- Contruct range
function AddrRange:AddrRange(start, stop)
	self.start = start
	self.stop = stop or start
end

-- Get addresses count
function AddrRange:getCount()
	return self.stop - self.start + 1
end

-- Properties
function AddrRange:__index(key)
	if key == 'count' then return self:getCount() end
	return getmetatable(self)[key]
end


-----------------------------------------------------------------------------
-- ModBus message
class "ModBusMessage"

-- ModBus message constructor (internally used in ModBus class)
function ModBusMessage:ModBusMessage(args)
	self.id = args.id
	self.callback = args.callback
	self.data = ''
	self.unit = args.unit or 0
	self.len = args.len or 0
	self.ok = false
	self.error = nil
	self.sent = args.sent or ''
	self.repeatCount = 1
	--ALERT('YYY:', 'new ModBusMessage', self.callback)  -- XXX
end

-- Unpack received data
function ModBusMessage:unpack(fmt)
	return self.ok and struct.unpack(fmt, self.data) or nil
end

-- Error message
function ModBusMessage.errorString(error)
	if type(error) ~= number and error ~= nil then error = error.error end
	return ModBus.errorString(error)
end

-- Burst mode for sending messages. Integer or nil if dissabled
ModBusMessage.burst = nil


-----------------------------------------------------------------------------
-- Sender helper for ModBus devices
class "ModBusIo"

local ioConnections = {}

function ModBusIo:ModBusIo(args)
	self.server = args.server
	self.port = args.port or 502
	self.id = args.id or (self.server .. tostring(self.port))
	self.sendQueue = {}
	self.messages = {}
	self.busy = false
	self.timeoutTimer = c8y.timer.new(REQUEST_TIMEOUT, function()
		self:busyOff()
		self:xdebug('timed out')
		self:reconnect(ERROR_TIMEOUT)
	end)
	self.timeoutTimer.singleShot = true
	self.burstTimer = not ModBusMessage.burst or (ModBusMessage.burst > 0 and c8y.timer.new(ModBusMessage.burst, function()
		self:busyOff()
		self:sendFromQueue()
		end) or false)
	if ModBusMessage.burst and ModBusMessage.burst > 0 then
		self.burstTimer.interval = ModBusMessage.burst
	end
	XDEBUG(string.format('ModBusIo(%s:%d) created, burst: %s', self.id, self.port, ModBusMessage.burst or 'none'))
end

-- Log object info
function ModBusIo:repr()
	return 'ModBusIo('..self.server..':'..self.port..'):'
end

-- WARN for this ModBusIO object (with title)
function ModBusIo:warning(...)
	c8y.log:stackLevelIncrement()
	WARN(self:repr(), ...)
end

-- INFO for this ModBusIO object (with title)
function ModBusIo:info(...)
	c8y.log:stackLevelIncrement()
	INFO(self:repr(), ...)
end

-- DEBUG for this ModBusIO object (with title)
function ModBusIo:debug(...)
	c8y.log:stackLevelIncrement()
	DEBUG(self:repr(), ...)
end

-- XDEBUG for this ModBusIO object (with title)
function ModBusIo:xdebug(...)
	c8y.log:stackLevelIncrement()
	XDEBUG(self:repr(), ...)
end

-- switch the busy timer off
function ModBusIo:busyOff()
	if self.busy and self.busy ~= true then
		self.busy:stop()
	end
	self.busy = false
end

-- send from Queue
function ModBusIo:sendFromQueue()
	if self.busy or next(self.sendQueue) == nil then return end
	if self.conn and self.conn.connected then
		self.busy = self.burstTimer
		if self.busy and self.busy ~= true then self.busy:start() end
		self.timeoutTimer:start(REQUEST_TIMEOUT)
		local msg = table.remove(self.sendQueue, 1)
		self:xdebug('Sending msg '..msg.id..':\n'..hex_dump(msg.sent))
		local conn = self:connection()
		if conn:send(msg.sent) then 
			self.messages[msg.id] = msg
		else
			-- reconnect
			table.insert(self.sendQueue, 1, msg)
			self:xdebug('send failed')
			self:reconnect()
		end
		--self:xdebug('Sent', id, msg)
	elseif not self.conn or not self.conn.running then
		self:xdebug(not self.conn and 'no connection' or self.conn.stateName)
		self:reconnect()
	end
end

-- data to send
function ModBusIo:send(msg)
	table.insert(self.sendQueue, msg)
	msg.timer = c8y.timer.singleShot(REQUEST_TIMEOUT + CONNECT_TIMEOUT, function()
		self:xdebug('message '..msg.id..' timed out')
		for i,m in ipairs(self.sendQueue) do
			if m.id == msg.id then table.remove(self.sendQueue, i) break end
		end
		self.messages[msg.id] = nil
		msg.ok = false
		msg.error = ERROR_TIMEOUT
		if msg.timer then msg.timer:stop() msg.timer = nil end
		if msg.callback ~= nil then msg.callback(msg) end
		self:sendFromQueue()
	end)
	self:sendFromQueue()
end

-- receive
function ModBusIo:receive()
	-- no connectiom nothing to receive
	if not self.conn then return end
	if self.conn.closed then self.conn = nil return end
	-- if we are inside of receiveing message, then receice data for this message
	if self.currentId then
		local msg = self.messages[self.currentId]
		if not msg then
			self:warning('Current message lost')
			self:restart()
			return
		end
		r = self.conn:recv{size = msg.len - #msg.data}   -- header (reqId, proto, length, unit)
		--self:xdebug('RecvData', r)
		if r and #r > 0 then
			self:xdebug('Data recevied:\n'..hex_dump(r))
			-- collect data
			msg.data = msg.data .. r
			-- all data received?
			if #msg.data == msg.len then
				self.currentId = nil  -- finished
				self.timeoutTimer:stop()
				self:busyOff()
				if #msg.data and msg.data:byte() < 0x80 then
					msg.ok = true
				else
					msg.ok = false
					msg.error = msg.data:byte(2)
					if not msg.error or msg.error == 0 then msg.error = ERROR_CONNECTION end
				end
				if msg.timer then msg.timer:stop() msg.timer = nil end
				--ALERT('YYY: read finihed', msg.callback)  -- XXX
				if msg.callback ~= nil then msg.callback(msg) end
				self.messages[msg.id] = nil
				self:sendFromQueue()
				return msg
			end
		end
	else
		-- if not inside message, then try receive header of new message
		r = self.conn:recv{size = 7}   -- header (reqId, proto, length, unit)
		--self:xdebug('RecvHead', r)
		if r and #r > 0 then
			self:xdebug('Header received:\n'..hex_dump(r))
			if #r < 7 then
				self:info('Incomplete response')
				self:reconnect(ERROR_DATA)
				return
			end
			local id, proto, len, unit = struct.unpack('!HHHB', r)
			local msg = self.messages[id]
			-- message must exists
			if msg == nil then
				self:warning('Unit:'..tostring(unit)..': Unknown response '..tostring(id))
				self:reconnect(ERROR_DATA)
				return
			end
			--  len - 1, because unit is inside length section and we receive it already
			msg.len, msg.unit = len - 1, unit
			-- now we inside message receiving
			self.currentId = id
			-- try receive data now
			return self:receive()
		end
	end
end


-- return connection for operations
function ModBusIo:connection()
	if self.conn == nil then
		conn = c8y.net:tcp()
		local host = self.server or c8y.config:value('server', '127.0.0.1')
		local port = self.port or c8y.config:value('port', 502)
		print('Connecting to '..host..':'..port)
		conn.timeout = self.__connectionTimeout or CONNECT_TIMEOUT
		self.conn = conn
		conn.onReceive = function()
			self:receive()
		end
		conn.onFinish = function()
			self.timeoutTimer:stop()
			if self.conn.error then
				self:flushMessages()
				self.timeoutTimer:start(RECONNECT_INTERVAL)
			else
				self:reconnect()
			end
		end
		conn.onConnect = function()
			self:xdebug('connected')
			self.error = nil
			self:sendFromQueue()
		end
		conn.onDisconnect = function()
			self:xdebug('disconnected, error='..self.conn.error..' ('..self.conn.errorMessage..')')
			self.data = ''
		end
		conn.onError = function()
			-- print only first error (not errors wall)
			if not self.error then
				self:warning('error='..self.conn.error..' ('..self.conn.errorMessage..')')
				self.error = self.conn.error
			end
		end
		local ok = conn:connect(host, port)
		if not ok then self:warning('Connection: '..err) end
	end
	return self.conn
end

-- fail messages
function ModBusIo:flushMessages(err)
	local messages = self.messages
	self.messages = {}
	self.currentId = nil
	local conn = self:connection()
	for _,msg in pairs(messages) do
		if msg.timer then msg.timer:stop() msg.timer = nil end
		if not msg.ok and (not err or err == 0) then err = ERROR_CONNECTION end
		if err then
			msg.ok = false
			msg.error = err
		end
		if msg.callback then msg.callback(msg) end
	end
end

-- force reconnect
function ModBusIo:reconnect(err)
	self:xdebug('reconnect ' .. (self.conn and 'active' or ''))
	--print(debug.traceback())
	if self.conn then
		self.conn.onFinish = nil
		self.conn:close()
	end
	self:busyOff()
	self.timeoutTimer:stop()
	self.conn = nil
	self.data = ''
	self:flushMessages(err)
	return conn
end

-- set connection timeout
function ModBusIo:setConnectionTimeout(timeout)
	self.__connectionTimeout = timeout
	if self.conn then self.conn.timeout = timeout end
end

-- close all connections
function ModBusIo.closeAll()
	for _,c in pairs(ioConnections) do
	if c.conn then
		c.conn.onFinish = nil
		c.conn:close()
	end
	end
end


-----------------------------------------------------------------------------
class "ModBus"

local ModBus_reqId = 1

-- Constructor, define ModBus object
-- args:
--   server  server to connect
--   port    port to conenct (TCP), default 502
function ModBus:ModBus(args)
	self.server = args.server or c8y.config:value('server', '127.0.0.1')
	self.port = args.port or c8y.config:value('port', 502)
	--self._reqId = 1   use ModBus_reqId
	self.proto = 0x0000                -- MODBUS protocol
	self.unit = args.unit or 0xff      -- Identification of a remote slave connected on a serial line or on other buses
	self.deviceId = args.deviceId
	self.name = args.name or ''
	local ioid = self.server .. ':' .. tostring(self.port) 
	if not ioConnections[ioid] then ioConnections[ioid] = ModBusIo{id=ioid, server=self.server, port=self.port} end
	self.io = ioConnections[ioid]
end

-- XDEBUG for this ModBus object (with title)
function ModBus:xdebug(...)
	c8y.log:stackLevelIncrement()
	XDEBUG('ModBus('..self.name..'):', ...)
end

-- MedBus properies
function ModBus:__index(key)
	-- true if same messages precessing
	if key == 'waiting' then return next(self.messages) ~= nil end
	return getmetatable(self)[key]
end

-- Add TCP ModBus layer (header)
function ModBus:addHeader(msg)
	local id = ModBus_reqId
	ModBus_reqId = (ModBus_reqId + 1) % 0x10000  -- 2 bytes ID
	return  struct.pack('!HHHB', id, self.proto, #msg + 1, self.unit or 0xff) .. msg, id
end

-- set connection timeout
function ModBus:setConnectionTimeout(timeout)
	self.io:setConnectionTimeout(timeout)
end

-- send
function ModBus:send(msg, callback)
	local msg, id = self:addHeader(msg)
	msg = ModBusMessage{id=id, callback=callback, unit=self.unit, sent=msg, modbus=self}
	self:xdebug('Send msg '..id)
	self.io:send(msg)
	return msg
end

-- receive
-- function ModBus:receive()
-- 	return self.io:receive()
-- end

-- Error message
function ModBus.errorString(error, x)
	print(debug.traceback())
	if type(error) ~= 'number' and x then  error = x  end
	if error ==  0 or error == nil then return 'No error' end
	if error ==  1 then return 'Illegal function' end
	if error ==  2 then return 'Illegal data address' end
	if error ==  3 then return 'Illegal data value' end
	if error ==  4 then return 'Slave device failure' end
	if error ==  5 then return 'Acknowledge' end
	if error ==  6 then return 'Slave device busy' end
	if error ==  8 then return 'Memory parity error' end
	if error == 10 then return 'Gateway path unavailable' end
	if error == 11 then return 'Gateway target device failed to respond' end
	if error == ERROR_TIMEOUT     then return 'Timed out' end
	if error == ERROR_DATA        then return 'Inconsistent data' end
	if error == ERROR_CONNECTION  then return 'Connection issue' end
	return 'Unknown error ('..tostring(error)..')'
end

-- get addresses ranges
function ModBus:_addrRanges(addr, limit)
	table.sort(addr)
	local seq = {}   -- seq[]  - list of addr ranges (start, end)
	for _,a in ipairs(addr) do
		if #seq > 0 and seq[#seq].stop + 1 == a and seq[#seq].count < limit then seq[#seq].stop = a
		else seq[#seq+1] = AddrRange(a)
		end
	end
	return seq
end

-- Milti-get of coils / registers
-- Callbacks:
--  - onSuccess(values, addr)   - called when operation is finished successfully
--  - onError(error, addr)      - called when error occurred
--  - onFinish(values, addr)    - always called when finihed, when error "values" is nil
--  - onProgress(n, N, addr)    - progress n of N (0/0 now)
--  "addr" is oryginal addresses table
function ModBus:_multiRead(args)
	local addr = args.addr
	local readFunc = args.readFunc
	local decodeFunc = args.decodeFunc
	local limit = args.limit or 64
	local values = args.values or {}
	local seq = {}
	if type(addr) == 'number' then
		self:xdebug('Read coils/registers, addr='..tostring(addr)..', count='..tostring(count))
		local count = args.count or 1
		--self:xdebug('MultiRead, count='..count, type(count))
		for a = 1,count,limit do
			seq[#seq+1] = AddrRange(a+addr-1, math.min(a+addr+limit-2, addr+count-1))
		end
	else
		seq = self:_addrRanges(addr, limit)
		self:xdebug('Read coils/registers, '..#addr..' adresses in '..#seq..' sequences')
	end
	local function cb(msg)
		--ALERT("YYY: _multiRead:callback", msg, msg and msg.err, args.addr, values)  -- XXX
		if msg then
			if msg.error then
				if args.onError ~= nil then args.onError(msg.error, args.addr) end
				if args.onFinish ~= nil then args.onFinish(nil, args.addr) end
				return false
			end
			decodeFunc(values, msg)
			--DEBUG('decoded', values)  -- XXX
		end
		if #seq > 0 then
			local range = table.remove(seq, 1)
			msg = readFunc(self, range.start, range.count, cb)
			--CRITICAL('Read', self.unit, string.byte(msg.sent:sub(8,8)), range.start, range.count, values)  -- XXX
			msg.range = range;
			if args.onProgress ~= nil then args.onProgress(0, 0, args.addr) end
			return true
		end
		if args.onSuccess ~= nil then args.onSuccess(values, args.addr) end
		if args.onFinish ~= nil then args.onFinish(values, args.addr) end
	end
	return cb()
end

-- load coil
function ModBus:_getCoil(modbusMsg, args)
	return self:_multiRead{addr=args.addr, count=args.count, onSuccess=args.onSuccess, onError=args.onError, onFinish=args.onFinish, onProgress=args.onProgress, values=args.values, limit=2000,
		readFunc=function(self, addr, n, callback)
			return self:send(struct.pack('!BHH', modbusMsg, addr, n), callback)
		end,
		decodeFunc=function(values, msg)
			--struct.unpack_bits_to_table{data=msg.data, from=3, count=msg.range.count, lsb=true, to=values, typeNumber=args.typeNumber}
			local list = {}
			struct.unpack_bits_to_table{data=msg.data, from=3, count=msg.range.count, lsb=true, to=list, typeNumber=args.typeNumber}
			for i=1,msg.range.count do
				values[i+msg.range.start-1] = list[i]
			end
		end
	}
end

-- load coil
function ModBus:getOutputCoil(args)
	return self:_getCoil(1, args)
end

-- input load coil
function ModBus:getInputCoil(args)
	return self:_getCoil(2, args)
end

-- save coil
function ModBus:setCoil(addr, val, callback)
	if type(val) == 'number' then val = val > 0 and true or false end
	return self:send(struct.pack('!BHH', 5, addr, (val and 0xff00 or 0)), callback)
end

-- (internal) load register worker
function ModBus:_getRegister(modbusMsg, args)
	return self:_multiRead{addr=args.addr, count=args.count, onSuccess=args.onSuccess, onError=args.onError, onFinish=args.onFinish, onProgress=args.onProgress, values=args.values, limit=125,
		readFunc=function(self, addr, n, callback)
			--ALERT('YYY:', callback)  -- XXX
			return self:send(struct.pack('!BHH', modbusMsg, addr, n), callback)
		end,
		decodeFunc=function(values, msg)
			--for i=1,msg.range.count do
			--	values[#values+1] = struct.unpack_hint(msg.data, false, false, 1+i*2)
			--end
			for i=1,msg.range.count do
				values[msg.range.start+i-1] = struct.unpack_hint(msg.data, false, false, 1+i*2)
			end
		end
	}
end

-- load input register
function ModBus:getInputRegister(args)
	return self:_getRegister(4, args)
end

-- load input register
function ModBus:getHoldingRegister(args)
	return self:_getRegister(3, args)
end

-- save register (single or multiple)
-- setRegister{data={...}, [addr=A,] [count=N,] onFinish=...} 
-- setRegister{addr=A, value=V, onFinish=...}
function ModBus:setRegister(args)
	local callback = args.onFinish
	local addr     = args.addr
	local data     = args.data
	local count    = args.count
	local value    = args.value
	if data then
		if not addr then
			for k,_ in pairs(data) do
				if addr == nil or k < addr then addr = k end
			end
		end
		if not count then
			count = 0
			while data[addr+count] do count = count + 1 end
		end
		if count < 1 then return end
		if count > 1 then
			local msg = struct.pack('!BHH', 16, addr, #value)
			for i = addr, addr + count - 1 do
				msg = msg .. struct.pack('!H', data[addr+i])
			end
			return self:send(msg, callback)
		end
		value = data[addr]
	end
	return self:send(struct.pack('!BHH', 6, addr, value), callback)
end


-----------------------------------------------------------------------------
-- API
modbusio = {}

modbusio.ModBus = ModBus
modbusio.ModBusMessage = ModBusMessage
modbusio.hex_dump = hex_dump

return modbusio
