
-----------------------------------------------------------------------------
-- http://www.modbus.org/docs/Modbus_Messaging_Implementation_Guide_V1_0b.pdf
-- http://www.modbus.org/docs/Modbus_Application_Protocol_V1_1b.pdf
-----------------------------------------------------------------------------
-- ModBus Plugin
--
-- All values (holding & input registers, coils and input bits) are in
-- one huge addressing space. More, every register bit is in one space too.
-- Every value (register or coil) has the same functions.
--
-- Address format (hex): NAAAAB, where
--  - N is namespace,
--  - A is register adress
--  - B is bit number (always 0 for coils)
-- 
--
-----------------------------------------------------------------------------

--local luatrace = require("luatrace")
--luatrace.tron{trace_file_name="/tmp/luatrace"}

require "tools"
require "bit"
BIT = bit.tobit
HEX = bit.tohex

--c8y.plugin.traceLine = true
--c8y.log.level = c8y.log.LOG_XDEBUG



AddrBitMask          = 0x00000f
AddrMask             = 0x0ffff0
AddrSpaceMask        = 0xf00000

AddrOutputCoils      = 0x000000  -- Discrete Output Coils
AddrInputCoils       = 0x100000  -- Discrete Input Contacts
AddrHoldingRegisters = 0x200000  -- Analog Output Holding Registers
AddrInputRegisters   = 0x300000  -- Analog Input Registers

UnsignedInt  = 0   -- up to 32-bit
SignedInt    = 1   -- up to 32-bit
Boolean      = 2   -- zero, non zero
Float        = 3   -- only for 32-bit values


-----------------------------------------------------------------------------
-- ModBus Measurements
class "MeasurementType"

function MeasurementType:MeasurementType(args)
	self.tmpl = args.tmpl
	-- data.changed = false
	-- self as standard table is uses: self[#self+1] = next_value_bit_address
end

-- Prepare measurement message (filled template)
function MeasurementType:message(args, data)
	local deviceId = args.deviceId or args[1] or 0
	local rec = { self.tmpl, deviceId }
	if args.now then rec[#rec + 1] = args.now end
	--ERROR('M', args.ifChanged, data and data.measurements)  -- XXX
	if args.ifChanged then
		local changed = false
		for _,v in ipairs(self) do
			local dvv = data.measurements[self.tmpl]
			--INFO(HEX(v.baddr), dvv, dvv and dvv[v.baddr], data.values[v.baddr])  -- XXX
			--NOTICE(HEX(v.baddr), dvv, dvv and dvv[v.baddr], data.values[v.baddr])  -- XXX
			if not dvv or dvv[v.baddr] ~= data.values[v.baddr] then changed = true break end
		end
		if not changed then return end
	end
	for _,v in ipairs(self) do
		if not data.measurements[self.tmpl] then data.measurements[self.tmpl] = {} end
		--WARN(HEX(v.baddr), data.values[v.baddr], '->', data.measurements[self.tmpl][v.baddr])
		data.measurements[self.tmpl][v.baddr] = data.values[v.baddr]
		if not data.values[v.baddr] then
			WARNING(string.format('No value %s for measurement %d', HEX(v.baddr,6), self.tmpl))
			XDEBUG('available values:', data.values)
			return
		end
		rec[#rec + 1] = data.values[v.baddr] * v.factor
	end
	--ERROR('M:', rec)
	return rec
end


-----------------------------------------------------------------------------
-- ModBus Alarm
class "AlarmType"

function AlarmType:AlarmType(args)
	self.tmpl = args.tmpl
	-- self.alarmType = args.alarmType
	-- self.severity = args.severity
	-- self.text = args.text
	-- data.id = args.id
end


-----------------------------------------------------------------------------
-- ModBus Event
class "EventType"

function EventType:EventType(args)
	self.tmpl = args.tmpl
end


-----------------------------------------------------------------------------
-- ModBus Value (register, coil)
class "ValueType"

-- Make bit-addr
function ValueType.makeAddr(args)
	if type(args) == 'number' then return args end
	if args.__type == 'Value' then return args.baddr end
	return BIT(args.baddr or (bit.lshift(args.addr,4) + (args.bit or 0) + (args.space or 0)))
end

-- Name of adress space
ValueType.addressShortName =
{
	[AddrOutputCoils      ] = 'Coil',              -- Discrete Output Coils
	[AddrInputCoils       ] = 'DiscreteInput',     -- Discrete Input Contacts
	[AddrHoldingRegisters ] = 'HoldingRegegister', -- Analog Output Holding Registers
	[AddrInputRegisters   ] = 'InputRegegister'    -- Analog Input Registers
}

-- Short name of adress space
ValueType.addressShortName =
{
	[AddrOutputCoils      ] = 'Coil',     -- Discrete Output Coils
	[AddrInputCoils       ] = 'In',       -- Discrete Input Contacts
	[AddrHoldingRegisters ] = 'HoldReg',  -- Analog Output Holding Registers
	[AddrInputRegisters   ] = 'InReg'     -- Analog Input Registers
}

-- Constructor
function ValueType:ValueType(args)
	self.baddr = Value.makeAddr(args)
	self.bsize = args.bsize or 1
	self.deviceType  = args.deviceType
	self.dataType    = args.dataType or UnsignedInt
	--data.value       = nil   -- current value
	self.status      = args.status
	self.alarm       = nil   -- AlarmType()
	self.event       = nil   -- EventType()
	self.measurement = nil   -- MeasurementType()
	mul = args.mul or 1
	div = args.div ~= 0 and args.div or 1
	self.factor = ( mul / div ) * math.pow(10, - (args.prec or 0))
	--WARN('New value', bit.tohex(self.baddr), self.bsize, self.factor)  -- XXX
end

-- Properties (simply only)
function ValueType:__index(key)
	if  key == 'addr'   then  return bit.rshift(bit.band(self.baddr, AddrMask), 4)    end
	if  key == 'bit'    then  return bit.band(self.baddr, AddrBitMask)                end
	if  key == 'count'  then  return 1 + bit.rshift(self.bit + self.bsize - 1, 4)     end
	if  key == 'space'  then  return bit.band(self.baddr, AddrSpaceMask)              end
	if  key == 'ashort' then  return ValueType.addressShortName[self.space] or '?'    end
	if  key == 'repr'   then  return self.ashort..'('..self.addr..':'..self.bit..')'  end
	return getmetatable(self)[key]
end

-- Read new value
function ValueType:read(args)
	local values = args.values or args
	local bbeg = self.bit
	local addr = self.addr
	local size = self.bsize
	local bsht = 0
	local val = 0
	while size > 0 do
		local mask = size < 32 and (bit.lshift(1, size) - 1) or BIT(0xffffffff)
		--ERROR(HEX(self.baddr), self.bit, self.bsize)  -- XXX
		--INFO(values)  -- XXX
		--NOTICE(addr, values[addr])  -- XXX
		val = val + bit.lshift(bit.band(bit.rshift(values[addr], bbeg), mask), bsht)
		addr, size, bbeg, bsht = addr + 1, size - (16 - bbeg), 0, bsht + (16 - bbeg)
	end
	local sign = bit.lshift(1,self.bsize-1)
	local range = self.bsize < 31 and bit.lshift(1,self.bsize) or self.bsize == 31 and 0x80000000 or 0x100000000
	if self.dataType == SignedInt then
		if self.bsize < 32 and bit.band(val, sign) ~= 0 then val = val - range end
	elseif self.dataType == UnsignedInt then
		if val < 0 then val = val + range end
	elseif self.dataType == Boolean then
		val = val ~= 0 and 1 or 0
	elseif self.dataType == Float then
		error "ModBus FLOAT data type is not supported yet"
	end
	return val
end

-- Update value (returns old one)
function ValueType:update(args, data)
	local newValue = self:read(args)
	local oldValue = data.values[self.baddr]
	data.values[self.baddr] = newValue
	return oldValue ~= newValue, newValue, oldValue
end

-- Change c8y value to binary value
function ValueType:_c8y_to_bin(val)
	val = self.factor == 0 and 0 or (val / self.factor)
	if self.dataType ~= Float then
		val = math.floor(val + 0.5)  -- round
	end
	return val
end

-- Set c8y value in given data array
function ValueType:fillValue(args)
	local data = args.data or {}
	local addr = self.addr
	local count = self.count
	--NOTICE(data, cont, addr)  -- XXX
	if               not data[addr + 0] then data[addr+0] = 0 end
	if count > 1 and not data[addr + 1] then data[addr+1] = 0 end
	if count > 2 and not data[addr + 2] then data[addr+2] = 0 end
	local start = self.bit
	local val   = args.value
	--NOTICE(data, start, val)  -- XXX
	val = self:_c8y_to_bin(val)
	val = bit.band(val, 0xffff)
	if start == 0 and self.bsize == 16 then  -- full register
		data[addr] = val
		return data
	end
	if start == 0 and self.bsize == 32 then  -- two full registers
		data[addr]   = bit.band(val, 0xffff)
		data[addr+1] = bit.rshift(val, 16)
		return data
	end
	-- not aligned bits...
	local mask  = self.bsize < 32 and (bit.lshift(1, self.bsize) - 1) or BIT(0xffffffff)
	local i = 0
	--ERROR(HEX(self.baddr), val, data) -- XXX
	while mask > 0 do
		local vm = bit.bnot(bit.band(bit.lshift(mask, start), 0xffff))
		local vv = bit.band(bit.lshift(val,  start), 0xffff)
		--NOTICE(i, start, HEX(mask), HEX(val), HEX(vm,4), HEX(vv,4)) -- XXX
		data[i+addr] = bit.bor(bit.band(data[i+addr], vm), vv)
		i, mask, val, start = i + 1, bit.rshift(mask, 16 - start), bit.rshift(val, 16 - start), 0
	end
	return data
end

-- Check if value is in range (fit into bits)
function ValueType:isInRange(args)
	--DEBUG(self, args)
	local val = self:_c8y_to_bin(args.value)
	local range = self.bsize < 31 and bit.lshift(1,self.bsize) or self.bsize == 31 and 0x80000000 or 0x100000000
	--CRITICAL('value', val, 'size', self.bsize, 'range', range, HEX(range))
	if self.dataType == UnsignedInt then
		return val >= 0 and val < range
	end
	if self.dataType == SignedInt then
		range = math.floor(range/2)
		return val < range  and val >= - range
	end
	if self.dataType == Boolean then
		return val == 0 or val == 1
	end
	return false
end

-----------------------------------------------------------------------------
-- Whole universe (all values in all spaces)
class "ValuesType"

function ValuesType:ValuesType(args)
	self.values = {}
	self.order = {}
	self.measurements = {}
end

-- Update space
-- Callbacks: onAlarm, onEvent, onNewMeasurement
function ValuesType:update(args, data)
	local universe = args.universe or args[1]
	--ERROR('update values', data.values, data.alarms)  -- XXX
	for _,v in ipairs(self.order) do
		local space = v.space
		local values = universe[space]
		if values then
			--XDEBUG('to update', HEX(space), values)  -- XXX
			local changed, new, old = v:update({values=values}, data)
			--WARN(' ... ', bit.tohex(v.baddr), changed, old, data.values, data.alarms)  -- XXX
			if changed then
				data.changed[space] = true
				if v.alarm and (old == nil or (new == 0) ~= (old == 0)) then
					if args.onAlarm then args.onAlarm(v, new * v.factor) end
				end
				--EMERG('event', bit.tohex(v.baddr), old, new, v.event) -- XXX
				if v.event and (old or not args.skipStartupEvents) then  -- skip events on startup
					if args.onEvent then args.onEvent(v, new * v.factor) end
				end
				if v.measurement then
					if args.onNewMeasurement then args.onNewMeasurement(v, new * v.factor) end
				end
			end
		end
	end
end

-- List of register addresses for given space 
function ValuesType:addresses()
	local universe = {}
	-- scan addresses
	for _,v in ipairs(self.order) do
		--NOTICE('var', bit.tohex(v.baddr))  -- XXX
		if not universe[v.space] then universe[v.space] = {} end
		local addrs = universe[v.space]
		local a = bit.rshift(bit.band(v.baddr, AddrMask), 4)
		if #addrs == 0 or addrs[#addrs] ~= a then
			--WARN(bit.tohex(v.baddr), v.count)  -- XXX
			for i = a, a + v.count - 1 do
				addrs[#addrs+1] = i
			end
		end
	end
	-- sort by addr
	-- for _,list in pairs(universe) do
	-- 	table.sort(list)
	-- end
	--ERROR('addresses:', universe)  -- XXX
	return universe
end

-- Get value object
function ValuesType:value(args)
	local baddr = ValueType.makeAddr(args)
	if not self.values[baddr] then
		if args.searchValue then  return  end
		self.values[baddr] = self:addValue(args)
		if not args.create then WARN('Unknown value', baddr) end
	end
	local value = self.values[baddr]
	return value
end

-- Add new value
function ValuesType:addValue(args)
	local value = args.__type == 'Value' and args or ValueType(args)
	if self.values[value.baddr] then WARN('Value already axists', value.baddr) end
	self.values[value.baddr] = value
	self.order[#self.order + 1] = value
	return value
end

-- Add status read/write
function ValuesType:addStatus(val)
	local value = self:value(val)
	value.status = true
end

-- Add new measurement
function ValuesType:addMeasurement(val, measurement)
	local value = self:value(val)
	local tmpl = measurement.tmpl
	if not self.measurements[tmpl] then self.measurements[tmpl] = MeasurementType{tmpl=tmpl} end
	local measurement = self.measurements[tmpl]
	measurement[#measurement + 1] = value
	value.measurement = tmpl
end

-- Finish initializating
function ValuesType:finish()
	table.sort(self.order, function(a,b) return a.baddr < b.baddr end)
	for _,m in pairs(self.measurements) do
		table.sort(m, function(a,b) return a.baddr < b.baddr end)
	end
end

-- Add new alarm
function ValuesType:addAlarm(val, alarm)
	local value = self:value(val)
	alarm = alarm.__type == 'Alarm' and alarm or AlarmType(alarm)
	value.alarm = alarm
end

-- Update Id of given alarm
function ValuesType:updateAlarmId(args, data)
	local value = self:value(args)
	local id  = args.id
	if value.alarm then
		data.alarms[value.baddr] = id
	end
end

-- Add new event
function ValuesType:addEvent(val, event)
	local value = self:value(val)
	event = event.__type == 'Event' and event or EventType(event)
	value.event = event
end

-- Space value iterator implementation
local function _Values_spaceIterator_next(op, current)
	--WARN('current', current, #op.values, bit.tohex(op.space))  -- XXX
	if current == 0 then
		-- this is good place to optimize finding begin of space (e.g. bin-tree)
		repeat
			current = current + 1
			local v = op.values.order[current]
			--DEBUG('ff', current, v and v.baddr, v and bit.tohex(v.baddr))  -- XXX
		until not v or v.space == op.space
	else
		current = current + 1
	end
	local v = op.values.order[current]
	--DEBUG('match?', current, v and v.baddr, v and bit.tohex(v.baddr))  -- XXX
	if v and v.space == op.space then return current, v end
end

-- Space value iterator
function ValuesType:spaceIterator(space)
	return _Values_spaceIterator_next, {values=self,space=space}, 0
end

-- Prepare status message (fill template data)
function ValuesType:message(args, data)
	local space   = args.space
	local message = args.message or {}
	--ERR(bit.tohex(space))  -- XXX
	for _,v in self:spaceIterator(space) do
		--NOTICE(bit.tohex(v.baddr), v.status)  -- XXX
		if v.status then
			--WARN(bit.tohex(v.baddr), data.values)  -- XXX
			message[#message + 1] = data.values[v.baddr] * v.factor
		end
	end
	return data
end


-----------------------------------------------------------------------------
-- ModBus Value (register, coil)
class "Value"

-- Make bit-addr
Value.makeAddr = ValueType.makeAddr

function Value:Value(args)
	self.type = args.type or args[1]
	self.value       = nil   -- current value
end

-- Read new value
function Value:read(data)
	return self.type:read(data)
end

-- Update value (returns old one)
function Value:update(args)
	local values = args.values or args[1]
	return type:update({values=values}, self)
end

-- C8y value (factored)
function Value:value()
	return self.value * self.type.factor, self
end

-- Value as ModBus binary field (not c8y value)
function Value:binaryValue()
	return self.value, self
end

-- Check if value is in range (fit into bits)
function Value:isInRange(args)
	return self.type:isInRange(args)
end

-----------------------------------------------------------------------------
-- Whole universe (all values in all spaces)
class "Values"

function Values:Values(args)
	self.type = args.type
	self.values = {}
	self.alarms = {}
	self.measurements = {}
	self.changed = {}
end

-- Get / create value
function Values:value(args)
	local baddr = ValueType.makeAddr(args)
	return baddr and self.values[baddr]
end

-- Update space
-- Callbacks: onAlarm, onEvent, onNewMeasurement
function Values:update(args)
	self.type:update(args, self)
end

-- Update Id of given alarm
function Values:updateAlarmId(args)
	self.type:updateAlarmId(args, self)
end

-- Get alarm ID
function Values:alarmId(args)
	local baddr = ValueType.makeAddr(args)
	return self.alarms[baddr]
end

-- List of register addresses for given space 
function Values:addresses()
	return self.type:addresses()
end

-- Prepare status message (fill template data)
function Values:message(args)
	return self.type:message(args, self)
end

-- Prepare measurement message (filled template)
function Values:measurementMessage(args)
	if args.measurement then
		return args.measurement:message(args, self)
	end
	local tmpl = args.tmpl
	return self.type.measurement[tmpl]:message(args, self)
end

-- Clear change flag(s)
function Values:clearChanges(args)
	local space = args.space or args
	if space == nil then
		self.changed = {}
	else
		if type(space) ~= 'table' then space = { space } end
		for _,s in ipairs(space) do
			self.changed[s] = nil
		end
	end
end

-----------------------------------------------------------------------------
-- API
modbusvalues = {}

modbusvalues.ValueType = ValueType
modbusvalues.ValuesType = ValuesType
modbusvalues.Value = Value
modbusvalues.Values = Values

modbusvalues.ADDRSPACE = {}
modbusvalues.ADDRSPACE.AddrBitMask          = AddrBitMask
modbusvalues.ADDRSPACE.AddrMask             = AddrMask
modbusvalues.ADDRSPACE.AddrSpaceMask        = AddrSpaceMask

modbusvalues.ADDRSPACE.AddrOutputCoils      = AddrOutputCoils
modbusvalues.ADDRSPACE.AddrInputCoils       = AddrInputCoils
modbusvalues.ADDRSPACE.AddrHoldingRegisters = AddrHoldingRegisters
modbusvalues.ADDRSPACE.AddrInputRegisters   = AddrInputRegisters

modbusvalues.DATATYPE = {}
modbusvalues.DATATYPE.UnsignedInt          = UnsignedInt
modbusvalues.DATATYPE.SignedInt            = SignedInt
modbusvalues.DATATYPE.Boolean              = Boolean
modbusvalues.DATATYPE.Float                = Float

return modbusvalues
