
-----------------------------------------------------------------------------
-- http://www.modbus.org/docs/Modbus_Messaging_Implementation_Guide_V1_0b.pdf
-- http://www.modbus.org/docs/Modbus_Application_Protocol_V1_1b.pdf
-----------------------------------------------------------------------------
-- ModBus Plugin
--
-- All values (holding & input registers, coils and input bits) are in
-- one huge addressing space. More, every register bit is in one space too.
-- Every value (register or coil) has the same functions.
--
-- Address format (hex): NAAAAB, where
--  - N is namespace,
--  - A is register adress
--  - B is bit number (always 0 for coils)
-- 
--
-----------------------------------------------------------------------------

--local luatrace = require("luatrace")
--luatrace.tron{trace_file_name="/tmp/luatrace"}

require "tools"
require "struct"
require "modbusio"
require "modbusvalues"
require "bit"
require "inifile"

ModBus = modbusio.ModBus
ModBusMessage = modbusio.ModBusMessage
ValuesType = modbusvalues.ValuesType
Values = modbusvalues.Values
ADDRSPACE = modbusvalues.ADDRSPACE
DATATYPE = modbusvalues.DATATYPE
BIT = bit.tobit
HEX = bit.tohex

--c8y.plugin.traceLine = true

local fieldbusUI = c8y.config:value('fieldbus', 4)
local sendMeasurementsOnlyIfChanged = value_to_boolean(c8y.config:value('measurementsOnChange', true))
local skipStartupEvents = value_to_boolean(c8y.config:value('skipStartupEvents', false))
local inifname = c8y.config:value('inifile', '/usr/local/etc/modbus.ini')
local ReadOnly = false  -- will be updated by ini file
ModBusMessage.burst = c8y.config:value('burst', 2)

-- some settings
local CONNECT_TIMEOUT = 15
local CoilsStatusTemplateId = 100
local RegistersStatusTemplateId = 101
local AlarmResponseTemplateId = 203

--c8y.log.level = c8y.log.LOG_XDEBUG

local function makeTmplAddrFrag(namespace, prefix)
	prefix = prefix or '$'
	local sbit = namespace == 'register' and ','..prefix..'.startBit' or ''
	if fieldbusUI == 2 then return prefix .. '.number'                               end
	if fieldbusUI == 3 then return prefix .. '.number'..sbit                         end
	if fieldbusUI == 4 then return prefix .. '.number'..sbit..','..prefix..'.input'  end
end

local TmplRegAddrFrag  = makeTmplAddrFrag('register')
local TmplCoilAddrFrag = makeTmplAddrFrag('coil')




-----------------------------------------------------------------------------
-- ModBus Device Type (device description)
class "DeviceType"

-- New device type
function DeviceType:DeviceType(typ)
	self.typ = typ
	self.id = typ:gsub('.*/','')
	self.id = tonumber(self.id)
	self.xid = self.id
	self.values = ValuesType()
	self.template = c8y.templates:create(self.xid)
	self.useServerTime = false
	self.updated = false
end

-- Send template with X-Id from this device type
function DeviceType:send(args)
	return self.template:send(args)
end

-- Add templates for discover device type
-- static function
function DeviceType.addTemplates(templates)
	templates.getTypeDevice = c8y.templates:addRequest('GET,/inventory/managedObjects/%%,,application/vnd.com.nsn.cumulocity.managedObject+json,%%,UNSIGNED,')
	templates.childDevice = c8y.templates:addResponse('$.childDevices.references,,$.managedObject.id,$.managedObject.name')
	templates.modbusDevice = c8y.templates:addResponse('$.c8y_ModbusDevice,,$.ipAddress,$.address,$.type')
	if fieldbusUI >= 4 then
		templates.modbusReadOnlyDevice = c8y.templates:addResponse(',,$.c8y_ModbusReadOnlyDevice')
		templates.updateReadOnly = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,,%%,UNSIGNED STRING,"{""c8y_ModbusReadOnlyDevice"": %%}"')
	end
	templates.modbusDeviceType = c8y.templates:addResponse(',$.c8y_useServerTime,$.c8y_useServerTime')
	templates.registersDef = c8y.templates:addResponse('$.c8y_Registers,,'..TmplRegAddrFrag..',$.multiplier,$.divisor,$.decimalPlaces'.. (fieldbusUI > 2 and ',$.noBits,$.signed' or ''))
	templates.regStatusDef = c8y.templates:addResponse('$.c8y_Registers,$.statusMapping,'..TmplRegAddrFrag)
	templates.coilStatusDef = c8y.templates:addResponse('$.c8y_Coils,$.statusMapping,'..TmplCoilAddrFrag)
	templates.measurementDef = c8y.templates:addResponse('$.c8y_Registers,$.measurementMapping,'..TmplRegAddrFrag..',$.measurementMapping.sendMeasurementTemplate')
	templates.coilAlaramsDef = c8y.templates:addResponse('$.c8y_Coils,$.alarmMapping,'..TmplCoilAddrFrag..',alarmMapping.raiseAlarmTemplate')
	templates.regAlaramsDef = c8y.templates:addResponse('$.c8y_Registers,$.alarmMapping,'..TmplRegAddrFrag..',alarmMapping.raiseAlarmTemplate')
	templates.coilEventsDef = c8y.templates:addResponse('$.c8y_Coils,$.eventMapping,'..TmplCoilAddrFrag..',eventMapping.eventTemplate')
	templates.regEventsDef = c8y.templates:addResponse('$.c8y_Registers,$.eventMapping,'..TmplRegAddrFrag..',eventMapping.eventTemplate')
end

-- Get register adress vector
-- returns record_offset, {address(number), start-bit, space}
function DeviceType.getRegAddr(rec, off)
	local input, output = ADDRSPACE.AddrInputRegisters, ADDRSPACE.AddrHoldingRegisters
	off = off or 3
	if fieldbusUI == 2 then return off+1, {addr=rec[off] - offset, bit=0, space=output} end
	if fieldbusUI == 3 then return off+2, {addr=rec[off] - offset, bit=rec[off+1], space=output} end
	return off+3, {addr=rec[off] - offset, bit=rec[off+1], space=value_to_boolean(rec[off+2]) and input or output}
end

-- Get coil adress vector
-- returns record_offset, {address(number), start-bit = 0, space}
function DeviceType.getCoilAddr(rec, off)
	local input, output = ADDRSPACE.AddrInputCoils, ADDRSPACE.AddrOutputCoils
	off = off or 3
	if fieldbusUI == 2 or fieldbusUI == 3 then return off+1, {addr=rec[off] - offset, bit=0, space=output} end
	return off+2, {addr=rec[off] - offset, bit=0, space=value_to_boolean(rec[off+1]) and input or output}
end

-- Update device type description
function DeviceType:update(templates)
	if not self.updated then
		local resp = templates.getTypeDevice:send{self.id}
		if resp.ok then
			--XDEBUG('Device', self.id, resp.records)
			for i = 1, resp.count do
				local rec = resp:get(i - 1)
				-- global settings
				if rec[1] == templates.modbusDeviceType.id then
					self.useServerTime = rec[3]
				end
				-- registers
				if rec[1] == templates.registersDef.id then
					local off, addr = DeviceType.getRegAddr(rec)
					local mul   = rec[off+0]
					local div   = rec[off+1]
					local prec  = rec[off+2]
					local size, sign = 0, DATATYPE.UnsignedInt
					if fieldbusUI >= 3 then
						size  = rec[off+3]
						sign  = value_to_boolean(rec[off+4]) and DATATYPE.SignedInt or DATATYPE.UnsignedInt
					end
					self.values:addValue{deviceType=self, addr=addr.addr, bit=addr.bit, space=addr.space, bsize=size,
					                     mul=mul, div=div, prec=prec, dataType=sign}
					XDEBUG('reg...', addr, rec)
				end
				-- registers - status
				if rec[1] == templates.regStatusDef.id then
					local off, addr = DeviceType.getRegAddr(rec)
					self.values:addStatus(addr)
				end
				-- registers - measurements
				if rec[1] == templates.measurementDef.id then
					local off, addr = DeviceType.getRegAddr(rec)
					local tmpl  = rec[off]
					self.values:addMeasurement(addr, {tmpl=tmpl})
				end
				-- registers - alarms
				if rec[1] == templates.regAlaramsDef.id then
					local off, addr = DeviceType.getRegAddr(rec)
					local tmpl  = rec[off]
					self.values:addAlarm(addr, {tmpl=tmpl})
				end
				-- registers - events
				if rec[1] == templates.regEventsDef.id then
					local off, addr = DeviceType.getRegAddr(rec)
					local tmpl  = rec[off]
					self.values:addEvent(addr, {tmpl=tmpl})
				end
				-- coils - status
				if rec[1] == templates.coilStatusDef.id then
					local off, addr = DeviceType.getCoilAddr(rec)
					addr.create = true
					addr.dataType = DATATYPE.Boolean
					self.values:addStatus(addr)
				end
				-- coils - alarms
				if rec[1] == templates.coilAlaramsDef.id then
					local off, addr = DeviceType.getCoilAddr(rec)
					local tmpl  = rec[off]
					addr.create = true
					addr.dataType = DATATYPE.Boolean
					self.values:addAlarm(addr, {tmpl=tmpl})
				end
				-- coils - events
				if rec[1] == templates.coilEventsDef.id then
					local off, addr = DeviceType.getCoilAddr(rec)
					local tmpl  = rec[off]
					addr.create = true
					addr.dataType = DATATYPE.Boolean
					self.values:addEvent(addr, {tmpl=tmpl})
				end
			end
			-- in coil/register number order  (templates: 100, 101)
			self.values:finish()
			-- Updated, skip next time
			self.updated = true
		end
	end
end


-----------------------------------------------------------------------------
-- Send message to cumulo
function ModBus:sendTemplate(args)
	return self.deviceType:send(args)
end


-----------------------------------------------------------------------------

-- Initialize plugin
function init()
	test()
	-- check if not read-only
	ini = inifile.parse(inifname)
	if ini and ini.modbus then ReadOnly = value_to_boolean(ini.modbus.readOnly) end
	--modbus = ModBus{}
	modbus = {}
	if fieldbusUI < 4 then
		c8y.operations:addOperation('c8y_ModbusDevice', '$.ipAddress,$.address,$.name,$.type', modbusDeviceOperation)
	else
		c8y.operations:addOperation('c8y_ModbusDevice', '$.ipAddress,$.address,$.name,$.type,$.id', modbusDeviceOperation)
	end
	c8y.operations:addOperation('c8y_ModbusDeviceRemoved', '$.childDeviceId', modbusDeviceRmovedOperation)
	if not ReadOnly then
		if fieldbusUI >= 3 then
			c8y.operations:addOperation('c8y_SetRegister', '$.value,$.address,$.register,$.startBit', setRegisterOperation)
		elseif fieldbusUI == 2 then
			c8y.operations:addOperation('c8y_SetRegister', '$.value,$.address,$.register', setRegisterOperation)
		end
		c8y.operations:addOperation('c8y_SetCoil', '$.value,$.address,$.coil', setCoilOperation)
	end
	c8y.operations:addOperation('c8y_ModbusConfiguration', '$.pollingRate,$.transmitRate', setModbusConfiguration)
	-- c8y_ModbusDeviceRemoved
	--devices = {}   - now in modbus table
	deviceTypes = {}
	templates={}
	templates.alarmResponse = c8y.templates:create(AlarmResponseTemplateId)
	templates.updateParentModbusConfiguration = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,application/vnd.com.nsn.cumulocity.managedObject+json,%%,UNSIGNED UNSIGNED UNSIGNED,"{""c8y_ModbusConfiguration"":{""protocol"": ""TCP"", ""pollingRate"": %%, ""transmitRate"": %%}}"')
	templates.update = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,application/vnd.com.nsn.cumulocity.managedObject+json,%%,UNSIGNED UNSIGNED STRING STRING,"{""c8y_ModbusDevice"":{""address"":%%,""type"":""%%"",""ipAddress"":""%%""}}"')
	templates.getDevice = c8y.templates:addRequest('GET,/inventory/managedObjects/%%,,application/vnd.com.nsn.cumulocity.managedObject+json,%%,UNSIGNED,')
	templates.getOperations = c8y.templates:addRequest('GET,/devicecontrol/operations?deviceId=%%&status=%%&pageSize=100,,,%%,UNSIGNED STRING,')
	templates.parentModbusResponse = c8y.templates:addResponse('$.c8y_ModbusConfiguration,,$.pollingRate,$.transmitRate')
	templates.opResponse = c8y.templates:addResponse('$.operations,,$.id,$.status,$.deviceId')
	templates.idResponse = c8y.templates:addResponse(',,$.id')
	templates.getUnresolvedAlarms = c8y.templates:addRequest('GET,/alarm/alarms?source=%%&status=%%,,application/vnd.com.nsn.cumulocity.alarmCollection+json,%%,UNSIGNED STRING,')
	templates.activeCoilAlarmResponse = c8y.templates:addResponse('$.alarms,$.c8y_ModbusSource,$.id,' .. makeTmplAddrFrag('coil', '$.c8y_ModbusSource.coil'))
	templates.activeRegAlarmResponse = c8y.templates:addResponse('$.alarms,$.c8y_ModbusSource,$.id,' .. makeTmplAddrFrag('register', '$.c8y_ModbusSource.register'))
  templates.createModbusAvailabilityAlarm = c8y.templates:addRequest('POST,/alarm/alarms,application/vnd.com.nsn.cumulocity.alarm+json,application/vnd.com.nsn.cumulocity.alarm+json,%%,UNSIGNED STRING STRING NOW,"{""source"":{""id"":""%%""},""type"":""%%"",""text"":""%%"",""severity"":""MAJOR"",""status"":""ACTIVE"",""time"":""%%""}"')
  templates.alarmCollectionIdAndTypeResponse = c8y.templates:addResponse("$.alarms,,$.id,$.type")
  
	DeviceType.addTemplates(templates)

	interval = {}
	interval.pollingRate = c8y.config:value('read_interval', 15)
	interval.transmitRate = c8y.config:value('measurement_interval', 30)
	interval.readTimeout = c8y.config:value('read_timeout', 30)
	offset = c8y.config:value('address_offset', 0)
end

-- Start this plugins (device integrated with c8y)
function start()
	scanModBusDevices()
	readRegisters()
	c8y.timer.singleShot(15, sendMeasurement)
end

-----------------------------------------------------------------------------

-- Create ModBus device
function createModBus(args)
	local m = ModBus(args)
	m.deviceType = args.deviceType
	m.values = Values{type=m.deviceType.values}
	modbus[args.deviceId] = m
	return m
end

-- Scan all ModBus children devices
function scanModBusDevices()
	INFO('Scanning ModBus devices')

	-- scan modbus intervals and children devices
	local hasModbusFragment = false
	local resp = templates.getDevice:send{c8y.device}
	if resp.ok then
		for i = 1, resp.count do
			local rec = resp:get(i - 1)
			if rec[1] == templates.parentModbusResponse.id then
				local pollingRate  = rec[3]
				local transmitRate = rec[4]
				hasModbusFragment = true
				interval.pollingRate = math.max(pollingRate, 1)
				interval.transmitRate = math.max(transmitRate, 1)
				if readRegistersTimer then readRegistersTimer.interval = interval.pollingRate end
				if sendMeasurementTimer then sendMeasurementTimer.interval = interval.transmitRate end
			elseif rec[1] == templates.childDevice.id then
				local devId   = rec[3]
				local devName = rec[4]
				local devresp = templates.getDevice:send{devId}
				if devresp.ok then
					for j = 1, devresp.count do
						local devrec = devresp:get(j - 1)
						if devrec[1] == templates.modbusDevice.id then
							local ip   = devrec[3]
							local unit = devrec[4]
							local typ  = devrec[5]
							DEBUG('Found ModBus device('..devId..'): '..ip..':'..unit..' named '..devName)
							if deviceTypes[typ] == nil then deviceTypes[typ] = DeviceType(typ) end
							createModBus{server=ip, unit=unit, deviceId=devId, name=devName, deviceType=deviceTypes[typ]}
						elseif templates.modbusReadOnlyDevice and devrec[1] == templates.modbusReadOnlyDevice.id then
							local ro = value_to_boolean(devrec[3])
							if ro ~= ReadOnly then
								templates.updateReadOnly:send{devId, ReadOnly and "true" or "false"}
							end
						end
					end
				end
				-- fail all outdated children set operations
				local opresp = templates.getOperations:send{devId,'PENDING'}
				if opresp.ok then
					for _,rec in pairs(opresp:getRecords(templates.opResponse.id)) do
						local id = rec[3]
						c8y.operations:setOperationStatus(id, c8y.operations.Executing)
					end
				end
				local opresp = templates.getOperations:send{devId,'EXECUTING'}
				if opresp.ok then
					for _,rec in pairs(opresp:getRecords(templates.opResponse.id)) do
						local id = rec[3]
						c8y.operations:setOperationStatus(id, c8y.operations.Failed, 'Operation outdated')
					end
				end
			end
		end
	end
	--DUMP('-x-x-x-x-x-x-x-x-x-x-x-x-   hasModbusFragment', hasModbusFragment, interval, 'tmplID', templates.parentModbusResponse.id)
	scanDeviceTypes()
	if not hasModbusFragment then
		templates.updateParentModbusConfiguration:send{c8y.device, interval.pollingRate, interval.transmitRate}
	end
	local unresolvedParentAlarmsResponse = templates.getUnresolvedAlarms:send{c8y.device, "ACTIVE"}
	local unresolvedModbusAvailabilityAlarms = {}
	if unresolvedParentAlarmsResponse.ok then
		for i=0, unresolvedParentAlarmsResponse.count - 1 do
			local record = unresolvedParentAlarmsResponse:get(i)
			if record[1] == templates.alarmCollectionIdAndTypeResponse.id then
				local id = record[3]
				local type = record[4]
				if string.match(type,"c8y_ModbusAvailability%d+Alarm") then
				  local key = string.match(type, "%d+", 5)
					unresolvedModbusAvailabilityAlarms[tonumber(key)] = id
				end
			end
		end
	end
	-- scan active alarams
	for _,m in pairs(modbus) do
		-- unresolved alarms
    if unresolvedModbusAvailabilityAlarms[m.unit] then
      m.availabilityAlarm = unresolvedModbusAvailabilityAlarms[m.unit]
      DEBUG("Detected unresolved modbus availability alarm for modbus " .. m.unit)
    end
		for _,status in pairs{'ACTIVE', 'ACKNOWLEDGED'} do
			resp = templates.getUnresolvedAlarms:send{m.deviceId, status}
			if resp.ok then
				for i = 1, resp.count do
					local rec = resp:get(i - 1)
					if rec[1] == templates.activeCoilAlarmResponse.id then
						local id = rec[3]
						local off, addr = DeviceType.getCoilAddr(rec, 4)
						addr.id = id
						m.values:updateAlarmId(addr)
						XDEBUG(string.format('Device(%s): Alarm %d (%s) for coil %d restored.', m.name, id, status, addr.addr))
					end
					if rec[1] == templates.activeRegAlarmResponse.id then
						local id = rec[3]
						local off, addr = DeviceType.getRegAddr(rec, 4)
						addr.id = id
						m.values:updateAlarmId(addr)
						XDEBUG(string.format('Device(%s): Alarm %d (%s) for register %d restored.', m.name, id, status, addr.addr))
					end
				end
			end
		end
	end
end

-- Scan all Device Types used by Modbus (child) devices
function scanDeviceTypes()
	for _, t in pairs(deviceTypes) do
		t:update(templates)
	end
end



-----------------------------------------------------------------------------
-- The c8y_SetRegister operation handler
function setModbusConfiguration(op)
	local rec = op.responseRecord
	local pollingRate  = rec[3]
	local transmitRate = rec[4]
	interval.pollingRate = math.max(pollingRate, 1)
	interval.transmitRate = math.max(transmitRate, 1)
	if readRegistersTimer then readRegistersTimer.interval = interval.pollingRate end
	if sendMeasurementTimer then sendMeasurementTimer.interval = interval.transmitRate end
	templates.updateParentModbusConfiguration:send{c8y.device, interval.pollingRate, interval.transmitRate}
	op.successful = true
	return true
end


-----------------------------------------------------------------------------
-- The c8y_ModbusDevice operation handler
function modbusDeviceOperation(op)
	if fieldbusUI >= 2 and fieldbusUI <= 4 then
		local record = op.responseRecord
		local ip   = record[3]
		local unit = record[4]
		local name = record[5]
		local typ  = record[6]
		local id  -- record[7] for fieldbus4, see below
		if fieldbusUI == 3 then
			-- force full scaning, fielbus3 created a child device but missed its ID
			DEBUG(string.format('ModBus child device %s found, but there is no ID, force full scanning...', name))
			scanModBusDevices()
		else
			if fieldbusUI == 2 then
				-- in fiedbus2 the agent creates a child device
				local device = c8y.device:childDevice(name, 'c8y_ModbusDevice', 'modbus:'..ip..':'..unit)
				XDEBUG('Adding child device', device.id)
				templates.update:send{device, unit, typ, ip}
				id = device.id
				DEBUG(string.format('ModBus child device %s (%s) added', name, id))
			else
				-- fieldbus4 creates a child device and resturns its ID
				id = record[7]
				DEBUG(string.format('ModBus child device %s (%s) found', name, id))
			end
			if deviceTypes[typ] == nil then deviceTypes[typ] = DeviceType(typ) end
			createModBus{server=ip, unit=unit, deviceId=id, name=name, deviceType=deviceTypes[typ]}
		end
	else
		op.setFailed('Unsupported fieldbus version: ' .. tostring(fieldbusUI))
		return
	end
	op.successful = true
	scanDeviceTypes()
	return true
end

-- ModBus device is removed in c8y
function modbusDeviceRmovedOperation(op)
	local record = op.responseRecord
	local id  = record[3]
	local m   = modbus[id]
	if m then
		modbus[id] = nil
		if m.modbus_reading then
			m.modbus_reading:stop()
			m.modbus_reading = nil
		end
		op.successful = true
	else
		op:Failed('No device ' .. tostring(id))
	end
end


-----------------------------------------------------------------------------
-- The c8y_SetCoil operation handler
function setCoilOperation(op)
	--DUMP('coilStatE:', op.id, op.name, op.records, op.deviceId)
	local m = modbus[op.deviceId]
	local record = op.responseRecord
	local val  = record[3]
	local unit = record[4]
	local addr = {searchValue=true, space=AddrOutputCoils}
	addr.addr  = record[5] - offset
	-- check arguments
	if not m then op:setFailed('No device ' .. tostring(op.deviceId)) return end
	if m.unit ~= unit then op:setFailed(string.format('Device %s: unit mismatch (%s != %s)', op.deviceId, m.unit, unit)) return end   -- ≠
	-- find value type
	local var = m.deviceType.values:value(addr)
	if not var then
		op:setFailed(string.format('No coil %s', addr.addr))
	elseif not var.status then
		op:setFailed('Coil ' .. var.repr .. ' has no status mapping')
	else
		m:setCoil(var.addr, val, function(msg)
				XDEBUG('set '..var.repr..' finished: op.id='..op.id..', modbus.id='..msg.id)
				if msg.ok then
					op.successful = true
				else
					op:setFailed('ModBus error: ' .. msg:errorString())
				end
			end)
	end
end


-----------------------------------------------------------------------------
-- The c8y_SetRegister operation handler
function setRegisterOperation(op)
	-- get arguments
	local m = modbus[op.deviceId]
	local record = op.responseRecord
	local val  = record[3]
	local unit = record[4]
	local addr = {searchValue=true, space=AddrHoldingRegisters}
	addr.addr = record[5] - offset
	if fieldbusUI >= 3 then
		addr.bit  = record[6]
	end
	-- check arguments
	if not m then op:setFailed('No device ' .. tostring(op.deviceId)) return end
	if m.unit ~= unit then op:setFailed(string.format('Device %s: unit mismatch (%s != %s)', op.deviceId, m.unit, unit)) return end   -- ≠
	-- find value type
	local var = m.deviceType.values:value(addr)
	if not var then
		op:setFailed(string.format('No register %s:%s', addr.addr, addr.bit))
	elseif not var.status then
		op:setFailed('Register ' .. var.repr .. ': no status mapping')
	elseif not var:isInRange{value=val} then
		op:setFailed('Register ' .. var.repr .. ': value out of range')
	else
		function finish(msg)
			XDEBUG('set '..var.repr..' finished: op.id='..op.id..', modbus.id='..msg.id)
			if msg.ok then
				op.successful = true
			else
				op:setFailed('ModBus error: ' .. msg:errorString())
			end
		end
		function setRegisters(vals)
			--CRITICAL('val', val, vals)  -- XXX
			var:fillValue{value=val, data=vals}
			m:setRegister{data=vals, onFinish=finish}
		end
		if var.bit == 0 and (var.bsize == 16 or var.bsize == 32) then  -- single, full register (or 2 full regs)
			setRegisters({})
		else
			local addr = var.addr
			if var.count > 1 then
				addr = {}
				for i = var.addr, var.addr + var.count - 1 do addrs[#addrs+1] = i end
			end
			--DEBUG('read to write', addr)  -- XXX
			m:getHoldingRegister
			{
				addr=addr,
				onError = function(err) op:setFailed('ModBus read error: ' .. ModBus.errorString(err)) end,
				onSuccess = setRegisters
			}
		end
	end
end


-----------------------------------------------------------------------------
-- Coils and registers read, measurement, alarms and events

-- Call read coils/registers
function readChain(args)
	local universe = args.universe
	local m = args.modbus
	local space, err
	local function call()
		if m.removed then return end
		if err then
			if args.onError  then args.onError(err) end
			if args.onFinish then args.onFinish(nil) end
			return
		end
		if m.modbus_reading then m.modbus_reading:start() end
		space = next(universe, space)
		if space then
			local fun
			if     space == ADDRSPACE.AddrOutputCoils      then fun = m.getOutputCoil
			elseif space == ADDRSPACE.AddrInputCoils       then fun = m.getInputCoil
			elseif space == ADDRSPACE.AddrInputRegisters   then fun = m.getInputRegister
			elseif space == ADDRSPACE.AddrHoldingRegisters then fun = m.getHoldingRegister
			end
			--CRITICAL(HEX(space), #universe[space])  -- XXX
			--WARN(universe[space])  -- XXX
			if fun then
				fun(m,  -- object self
						{   -- method agruments
							addr       = universe[space],
				      onError    = function(e) err = e end,
				      onSuccess  = function(values) universe[space] = values end,
				      onProgress = function() if m.modbus_reading then m.modbus_reading:start() end end,
				      onFinish   = call,
							typeNumber = true  -- use 0/1 in coils (not false/true)
				    })
			end
		else
			if args.onSuccess then args.onSuccess(universe) end
			if args.onFinish  then args.onFinish(universe) end
		end
	end
	call()
end

local function raiseModbusAvailabilityAlarm(m, errorMessage)
  local alarmType = string.format("c8y_ModbusAvailability%dAlarm", m.unit)
  local alarmText = string.format("Error connecting to modbus device %s (%s:%s) failed, %s",m.name, m.server, m.port, errorMessage)
  local response = templates.createModbusAvailabilityAlarm:send{c8y.device, alarmType, alarmText}
  if response.ok then
    for i=0, response.count - 1 do
      local record = response:get(i)
      if record[1] == templates.idResponse.id then
        m.availabilityAlarm = record[3]
      end
    end
  end
end

local function clearModbusAvailabilityAlarm(m)
	c8y.device:clearAlarm(m.availabilityAlarm)
	m.availabilityAlarm = nil
end


-- Reading registers
function readRegisters()
	if readRegistersTimer == nil then
		readRegistersTimer = c8y.timer.start(interval.pollingRate, readRegisters)
	end
	if next(modbus) == nil then return end
	XDEBUG('Updating coils and registers status. Interval '..readRegistersTimer.interval)
	for _,m in pairs(modbus) do
		if m.modbus_reading then
			XDEBUG('ModBus device' .. m.name .. ' updating coils and registers status in progress - ignoring')
		else
			local function failed(errmsg)
				local errmsg = string.format('ModBus %d: reading failed: %s', m.unit, type(errmsg)=='string' and errmsg or '')
        if m.availabilityAlarm then
          XDEBUG(string.format("Modbus availability alarm for unit %d already present skipping", m.unit))
        else
          XDEBUG(string.format("Raising modbus availability alarm for unit %d", m.unit))
          raiseModbusAvailabilityAlarm(m,errmsg)
        end
				if m.lastError == errmsg then
					DEBUG(errmsg)
				else
					WARNING(errmsg)
				end
				m.lastError = errmsg
				m.modbus_reading = nil
			end
			m.modbus_reading = c8y.timer.singleShot(interval.readTimeout, failed)
			-- all values to read (registers & coils)
			local universe = m.values:addresses()
			-- read them all
			readChain {
				universe  = universe,
				modbus    = m,
				onError   = failed,
				onSuccess = function()
					-- update values
					if m.modbus_reading then m.modbus_reading:stop() m.modbus_reading = nil end
					local now = (fieldbusUI <= 2 or not m.deviceType.useServerTime) and c8y.strftime() or nil
					--DEBUG('was read', universe)  -- XXX
          if m.availabilityAlarm then
            XDEBUG(string.format("Clearing modbus availability alarm for unit %d", m.unit))
            clearModbusAvailabilityAlarm(m)
          end
					m.values:update{
						universe=universe,
						skipStartupEvents=skipStartupEvents,   -- true for skipping events on start-up
						onAlarm = function(value, level)
							--NOTICE('Alarm!', value.repr, HEX(value.baddr), level)  -- XXX
							if level ~= 0 then
								local resp = m:sendTemplate{value.alarm.tmpl, m.deviceId, now}
								if resp.ok then
									local id = AlarmResponseTemplateId and resp:getCustomId(AlarmResponseTemplateId) or true
									m.values:updateAlarmId{baddr=value.baddr, id=id}
									XDEBUG(m.name, value.repr ..' alarm at '..(now or 'now')..': '..tostring(id))
								end
							else
								local id = m.values:alarmId{baddr=value.baddr}
								if id then
									c8y.Device(m.deviceId):clearAlarm(id)
									m.values:updateAlarmId{baddr=value.baddr, id=nil}
									XDEBUG(m.name, value.repr ..' alarm cleared')
								end
							end
						end,
						onEvent = function(value, level)
						  if now then
						  	m:sendTemplate{value.event.tmpl, m.deviceId, now, level}
						  else
						    m:sendTemplate{value.event.tmpl, m.deviceId, level}
						  end
						end,
					}
					m.deviceType.template:flush()
					-- send coils status
					if m.removed then return end
					if m.values.changed[ADDRSPACE.AddrOutputCoils] or m.values.changed[ADDRSPACE.AddrInputCoils] then
						local toSend = { CoilsStatusTemplateId, m.deviceId }
						m.values:message{space=ADDRSPACE.AddrOutputCoils, message=toSend}
						m.values:message{space=ADDRSPACE.AddrInputCoils, message=toSend}
						toSend.Collect = true
						m:sendTemplate(toSend)
						m.values:clearChanges{space={ADDRSPACE.AddrOutputCoils, ADDRSPACE.AddrInputCoils}}
					end
					-- send registers status
					if m.removed then return end
					if m.values.changed[ADDRSPACE.AddrHoldingRegisters] or m.values.changed[ADDRSPACE.AddrInputRegisters] then
						local toSend = { RegistersStatusTemplateId, m.deviceId }
						m.values:message{space=ADDRSPACE.AddrHoldingRegisters, message=toSend}
						m.values:message{space=ADDRSPACE.AddrInputRegisters, message=toSend}
						toSend.Collect = true
						m:sendTemplate(toSend)
						m.values:clearChanges{space={ADDRSPACE.AddrHoldingRegisters, ADDRSPACE.AddrInputRegisters}}
					end
					-- send it all
					m.deviceType.template:flush()
					m.lastError = nil
				end
			}
		end
	end
end

-- Send measurement
function sendMeasurement()
	if sendMeasurementTimer == nil then
		sendMeasurementTimer = c8y.timer.start(interval.transmitRate, sendMeasurement)
	end
	if next(modbus) == nil then return end
	XDEBUG('Send measurement. Interval '..sendMeasurementTimer.interval)
	for _,m in pairs(modbus) do
		local now = not m.deviceType.useServerTime and c8y.strftime() or nil
		for _,measurement in pairs(m.values.type.measurements) do
			local toSend = m.values:measurementMessage
			{
				measurement = measurement,
				deviceId    = m.deviceId,
				ifChanged   = sendMeasurementsOnlyIfChanged,
				now         = now
			}
			if toSend then
				toSend.Collect = true
				m:sendTemplate(toSend)
			end
		end
		m.deviceType.template:flush()
	end
end


-----------------------------------------------------------------------------
-- Stop plugin (close all connections)
function stop()
	modbus = {}
	deviceTypes = {}
	ModBusIo.closeAll()
end

-----------------------------------------------------------------------------


function test()
end

